/***************************************************************************
 *   Copyright (C) 2008, 2009, 2010 by Stephane List                       *
 *   slist@lilotux.net                                                     *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "plc.h"
#include "net_pcap.h"
#include <ctype.h>

#define VERSION "0.1.7"

enum {
	COLUMN_DEVICE_NAME = 0,
	COLUMN_RX_RATE,
	COLUMN_TX_RATE,
	COLUMN_MAC,
	COLUMN_LAST_DATE,
	COLUMN_LAST
};

#define AUTOSCAN_TEMPO  1 // scan every x seconds
#define AUTOSCAN_REMOVE (AUTOSCAN_TEMPO*2)+1 // remove old entries after x seconds of inactivity

plc::plc(QWidget * parent, Qt::WindowFlags f)
: QDialog(parent, f)
{
	int res = -1;
	if_count = 0;
	MyMac = "";
	autoscan = true;
	closing = false;
	pcapThread = NULL;

	QCoreApplication::setOrganizationName("Easyvisio");
	QCoreApplication::setOrganizationDomain("easyvisio.com");
	QCoreApplication::setApplicationName("plc");
	QSettings settings;

	QIcon icon = QIcon(":/plc.ico");
	setWindowIcon(icon);

	IfGroup = new QGroupBox(tr("Choose a network interface"));
	QVBoxLayout *IfLayout = new QVBoxLayout;

	QComboBox *combo = new QComboBox();

	Q_FOREACH (QNetworkInterface my_if, QNetworkInterface::allInterfaces()) {
		if ( !(my_if.flags() & QNetworkInterface::IsLoopBack) &&
				(my_if.flags() & QNetworkInterface::IsUp) &&
				(my_if.flags() & QNetworkInterface::IsRunning))
		{
			qDebug() << "Interface found:" << my_if.name();
			combo->addItem(my_if.name());
			if (if_count++ == 0)
			{
				/* We will take the first IF */
				res = plcnet_select_if(my_if);
				if (res < 0)
				{
					QMessageBox::warning(this, tr("plc"), tr("Network interface not supported !"), QMessageBox::Ok);
				}
				else
				{
					qDebug() << "On lance un thread de sniff.";
					pcapThread = new PcapThread(this);
					pcapThread->start(QThread::NormalPriority);
				}
			}
		}
	}
	if (if_count == 0)
	{
		QMessageBox::critical(this, tr("plc"), tr("No network interface found!"), QMessageBox::Ok);
		exit(0);
	}
	IfLayout->addWidget(combo);

	IfGroup->setLayout(IfLayout);
	if (res < 0)
	{
		QMessageBox::critical(this, tr("plc"), tr("Impossible to open socket, you need to be administrator!"), QMessageBox::Ok);
	}
#if defined (Q_OS_WIN32)
	setMinimumSize(700, 500); 
#else
	setMinimumSize(900, 600); 
#endif
	setWindowTitle(tr("plc"));
	createTab();

	qDebug("Chargement des anciens Devices vus dans plc");
	int i=0;
	MainTab_->ScannedTable->setSortingEnabled(false);
	foreach (QString mac, settings.childGroups()) {
		qDebug() << "Chargement des anciens Devices vus dans plc :" << mac << "(" << i <<")";
		MainTab_->ScannedTable->setRowCount(i+1);
		settings.beginGroup(mac);
		MainTab_->ScannedTable->setItem(i, COLUMN_DEVICE_NAME,new QTableWidgetItem(settings.value("dev_name"," ").toString()));
		MainTab_->ScannedTable->setItem(i, COLUMN_MAC,new QTableWidgetItem(mac));
		MainTab_->ScannedTable->setItem(i, COLUMN_LAST_DATE,new QTableWidgetItem(settings.value("last_date"," ").toString()));
		MainTab_->ScannedTable->item(i, COLUMN_MAC)->setFlags(MainTab_->ScannedTable->item(i, COLUMN_MAC)->flags() & ~Qt::ItemIsEditable);
		MainTab_->ScannedTable->item(i, COLUMN_LAST_DATE)->setFlags(MainTab_->ScannedTable->item(i, COLUMN_LAST_DATE)->flags() & ~Qt::ItemIsEditable);
		MainTab_->ScannedTable->resizeColumnToContents(COLUMN_MAC);
		i++;
		settings.endGroup();
	}
	MainTab_->ScannedTable->setSortingEnabled(true);
	qDebug("Chargement des anciens Devices vus dans plc (fini)");

	connect(this, SIGNAL(Mac_detected(const QString &)),                     MainTab_, SLOT(add_mac(const QString &)));
	connect(this, SIGNAL(My_Mac_detected(const QString &, const QString &)), MainTab_, SLOT(add_my_mac(const QString &, const QString &)));
	connect(this, SIGNAL(Rate_detected(const QString &, int, int)),          MainTab_, SLOT(add_rate(const QString &, int, int)));
	connect(this, SIGNAL(Date_detected(const QString &)),                    MainTab_, SLOT(add_date(const QString &)));
	connect(this, SIGNAL(Nek_changed_confirm()),                             PrivacyTab_, SLOT(display_nek_confirm()));

	scanThread = new ScanThread(this);
	connect(scanThread, SIGNAL(remove_old()),                          MainTab_,   SLOT(remove_old()));
	connect(combo,      SIGNAL(currentIndexChanged(const QString &)),  this,       SLOT(change_if(const QString &)));
	scanThread->start(QThread::NormalPriority);
}

void plc::createTab()
{
	tabWidget = new QTabWidget(this);
	MainTab_ = new MainTab(this);
	PrivacyTab_ = new PrivacyTab(this);

	tabWidget->addTab(MainTab_, tr("Main"));
	tabWidget->addTab(PrivacyTab_, tr("Privacy"));
	tabWidget->addTab(new AboutTab(this), tr("About"));

	QVBoxLayout *mainLayout = new QVBoxLayout(this);
	mainLayout->addWidget(tabWidget);
	if (if_count>1)
	{
		mainLayout->addWidget(IfGroup);
	}
	setLayout(mainLayout);
}

void plc::setMyMac(const QString& mac)
{
	MyMac = mac;
}

QString& plc::getMyMac()
{
	return MyMac;
}

void plc::closeEvent ( QCloseEvent * event )
{
	QTableWidget *ScannedTable = MainTab_->ScannedTable;
	QSettings settings;
	int row;

	qDebug("Closing, save history");
	closing = true;
	autoscan = false;
	disconnect(this, 0, 0, 0);
	if (pcapThread != NULL)
		disconnect(pcapThread, 0, 0, 0);

	if (scanThread != NULL)
		disconnect(scanThread, 0, 0, 0);

	for (row=0; row< (ScannedTable->rowCount()); row++)
	{
		settings.beginGroup(ScannedTable->item(row, COLUMN_MAC)->text());
		settings.setValue("dev_name" , ScannedTable->item(row, COLUMN_DEVICE_NAME)->text());
		settings.setValue("last_date", ScannedTable->item(row, COLUMN_LAST_DATE)->text());
		settings.endGroup();
	}
	settings.sync();

	if (scanThread != NULL)
	{
		if (! scanThread->wait(AUTOSCAN_TEMPO * 2000))
		{
			qDebug() << "Warning : scanThread not ended";
		}
	}
	if (pcapThread != NULL)
	{
		pcapThread->terminate();
		if (! pcapThread->wait(AUTOSCAN_TEMPO * 2000))
		{
			qDebug("Warning : pcapThread not ended");
		}
		else
		{
			plcnet_deinit();
		}
	}
	event->accept();
}

MainTab::MainTab(plc *parent) : QTabWidget(parent)
{
	int h;

	Parent = parent;

	QLabel *Label1 = new QLabel(tr("Local Device on your computer:"));

	ConnectedTable = new QTableWidget(1,2);
	QStringList htab;
	htab << tr("MAC Address") << tr("Firmware");
	ConnectedTable->setHorizontalHeaderLabels(htab);
#if QT_VERSION >= 0x040300
	ConnectedTable->setCornerButtonEnabled(false);
#endif
	ConnectedTable->resizeColumnsToContents();
	ConnectedTable->setRowCount(1);

	h = ConnectedTable->rowHeight(0);
	h += ConnectedTable->horizontalHeader()->height();
	ConnectedTable->setMaximumHeight(h);
	ConnectedTable->setVerticalScrollBarPolicy(Qt::ScrollBarAlwaysOff);
	ConnectedTable->horizontalHeader()->setStretchLastSection(true);
	ConnectedTable->setEditTriggers(QAbstractItemView::NoEditTriggers);

	QLabel *Label2 = new QLabel(tr("Device(s) detected:"));

	ScannedTable = new QTableWidget();
	ScannedTable->setMinimumSize(0, 400);
	ScannedTable->setRowCount(0);
	ScannedTable->setColumnCount(COLUMN_LAST);
	//  ScannedTable->setEditTriggers(QAbstractItemView::NoEditTriggers);
	QStringList htab2;
	htab2 << tr("Device Name")
		//	<< tr("Password")
		//	<< tr("Quality")
		<< tr("RX Rate(Mbps)")
		<< tr("TX Rate(Mbps)")
		<< tr("MAC Address")
		<< tr("Last Date");
	ScannedTable->setHorizontalHeaderLabels(htab2);
	ScannedTable->horizontalHeader()->setStretchLastSection(true);

	ScannedTable->sortItems(0, Qt::AscendingOrder);
	ScannedTable->setSortingEnabled(true);
#if QT_VERSION >= 0x040300
	ScannedTable->setCornerButtonEnabled(false);
#endif
	ScannedTable->resizeColumnsToContents();

	QVBoxLayout *mainLayout = new QVBoxLayout;
	mainLayout->addWidget(Label1);
	mainLayout->addWidget(ConnectedTable);
	mainLayout->addWidget(Label2);
	mainLayout->addWidget(ScannedTable);
	mainLayout->addStretch(1);
	setLayout(mainLayout);
}

void MainTab::add_mac(const QString& mac)
{
	QString& my = Parent->getMyMac();

	if (Parent->closing)
		return;

	qDebug() << "add MAC :" << mac;

	QList<QTableWidgetItem*>MyQList = ScannedTable->findItems(mac, Qt::MatchFixedString);
	if (MyQList.size() == 0)
	{
		if ((! my.isEmpty()) && (my != mac))
		{
			return;
		}
		else
		{
			if (my.isEmpty())
			{
				plc_req_version(mac);
			}
		}
	}
}

void MainTab::add_my_mac(const QString &mac, const QString &firmware)
{
	QList<QTableWidgetItem*>MyQList = ConnectedTable->findItems(mac, Qt::MatchFixedString);

	qDebug() << "add_my_mac start";

	if (MyQList.size() == 0)
	{
		int i = ConnectedTable->rowCount();
		ConnectedTable->insertRow(i);
		ConnectedTable->setItem(0,0,new QTableWidgetItem(mac));
		ConnectedTable->setItem(0,1,new QTableWidgetItem(firmware));
		ConnectedTable->resizeColumnToContents(0);
		Parent->setMyMac(mac);
	}
	qDebug() << "add_my_mac end";
}

/*
 * mac = MAC address
 * rate = rate in Mbit/s
 * direction = 0 for RX, 1 for TX */
void MainTab::add_rate(const QString &mac, int rate, int direction)
{
	QString my = Parent->getMyMac();
	int col, i;
	QList<QTableWidgetItem*>MyQList = ScannedTable->findItems(mac, Qt::MatchFixedString);

	if (( my.isEmpty()) || (my == mac))
		return;

	if (direction == 0)
	{
		col = COLUMN_RX_RATE;
	}
	else
	{
		col = COLUMN_TX_RATE;
	}
	if (MyQList.size() == 0)
	{
		i = ScannedTable->rowCount();
		ScannedTable->setSortingEnabled(false);
		ScannedTable->setRowCount(i+1);

		QTableWidgetItem *item0 = new QTableWidgetItem(QString("Device %1").arg(i+1));
		ScannedTable->setItem(i, COLUMN_DEVICE_NAME, item0);

		ScannedTable->setItem(i, COLUMN_MAC,new QTableWidgetItem(mac));
		ScannedTable->item(i,COLUMN_MAC)->setFlags(ScannedTable->item(i, COLUMN_MAC)->flags() & ~Qt::ItemIsEditable);

		ScannedTable->setItem(i,col,new QTableWidgetItem(QString("%1").arg(rate)));
		ScannedTable->item(i,col)->setFlags(ScannedTable->item(i, col)->flags() & ~Qt::ItemIsEditable);

		(ScannedTable->item(i,col))->setTextAlignment(Qt::AlignCenter);
		ScannedTable->resizeColumnToContents(COLUMN_DEVICE_NAME);
		ScannedTable->resizeColumnToContents(col);
		ScannedTable->resizeColumnToContents(COLUMN_MAC);
	}
	else /* We need to update the entry */
	{
		i = MyQList[0]->row();
		ScannedTable->setItem(i,col,new QTableWidgetItem(QString("%1").arg(rate)));
		(ScannedTable->item(i,col))->setTextAlignment(Qt::AlignCenter);
		ScannedTable->item(i,col)->setFlags(ScannedTable->item(i, col)->flags() & ~Qt::ItemIsEditable);
		ScannedTable->resizeColumnToContents(col);
	}
}

// Update Last seen value
void MainTab::add_date(const QString &mac)
{
	int row;

	QList<QTableWidgetItem*>MyQList = ScannedTable->findItems(mac, Qt::MatchFixedString);
	if (MyQList.size() != 0)
	{
		QDateTime now = QDateTime::currentDateTime();
		row = MyQList[0]->row();
		ScannedTable->setItem(row, COLUMN_LAST_DATE,new QTableWidgetItem(now.toString("dddd d/M/yyyy hh:mm:ss")));
		ScannedTable->item(row, COLUMN_LAST_DATE)->setFlags(ScannedTable->item(row, COLUMN_LAST_DATE)->flags() & ~Qt::ItemIsEditable);
	}
}

void MainTab::remove_old()
{
	int row;

	QDateTime now = QDateTime::currentDateTime();

	for (row=0; row < (ScannedTable->rowCount()); row++)
	{
		if ((ScannedTable->item(row, COLUMN_LAST_DATE) != NULL) && (ScannedTable->item(row, COLUMN_LAST_DATE)->text().isEmpty() != true))
		{
			QDateTime last = QDateTime::fromString(ScannedTable->item(row, COLUMN_LAST_DATE)->text(), "dddd d/M/yyyy hh:mm:ss");
			if (last.secsTo(now) > AUTOSCAN_REMOVE)
			{
				ScannedTable->item(row, COLUMN_LAST_DATE)->setBackgroundColor(Qt::red);
				//				qDebug("remove_old (%d seconds)!!!", last.secsTo(now));
			}
		}
	}
}

PrivacyTab::PrivacyTab(plc *parent) : QTabWidget(parent)
{
	Parent = parent;
	Nek_state = NEK_IDLE;
	QLabel *Label1 = new QLabel(tr("Use this screen to create a Private Network that provides you with extra security.\n"));

	QGroupBox *Group1 = new QGroupBox(tr("Private Network Name"));
	nekLineEdit = new QLineEdit;
	nekLineEdit->setText(QString("HomePlug"));
	nekLineEdit->setFocus();
	QLabel *orLabel = new QLabel(tr("OR"));
	QPushButton *pb1 = new QPushButton(tr("Use Default (Public Network)"), this);
	connect(pb1, SIGNAL(clicked()), this, SLOT(setDefaultNek()));

	QLabel *warnLabel = new QLabel(tr("Do not share the network name with others you do not want to be part of this network."));
	QGridLayout *nekLayout = new QGridLayout;
	nekLayout->addWidget(nekLineEdit, 0, 0);
	nekLayout->addWidget(orLabel, 0, 1);
	nekLayout->addWidget(pb1, 0, 2);
	nekLayout->addItem(new QSpacerItem(1,1,QSizePolicy::Expanding,QSizePolicy::Minimum), 0, 3);
	nekLayout->addWidget(warnLabel, 1, 0, 1, 4, 0);
	Group1->setLayout(nekLayout);

	QLabel *Label2 = new QLabel(tr("\nAfter setting the name above, choose how it will be applied below:\n"));

	QGroupBox *Group2 = new QGroupBox(tr("Place On This Network"));

	QLabel *Label3 = new QLabel(tr("Only the device attached to this computer (the Local Device),\n(Isolates this computer from others)"));
	Label3->setAlignment(Qt::AlignRight|Qt::AlignVCenter);

	QLabel *Label4 = new QLabel(tr("ALL devices whose Password has been entered.\n(They communicate together but are isolated from devices with a different network name)" ));
	Label4->setAlignment(Qt::AlignRight|Qt::AlignVCenter);
	Label4->setWordWrap(true);

	QPushButton *pb2 = new QPushButton(tr("Set Local Device Only"), this);
	connect(pb2, SIGNAL(clicked()), this, SLOT(setLocalDevice()));

	QLabel *Label5 = new QLabel(tr("OR" ));
	Label5->setAlignment(Qt::AlignHCenter|Qt::AlignVCenter);

	QPushButton *pb3 = new QPushButton(tr("Set All Devices (buggy, so disabled :-< )"), this);
	pb3->setEnabled(false);
	connect(pb3, SIGNAL(clicked()), this, SLOT(setAllDevices()));

	QGridLayout *placeLayout = new QGridLayout;
	placeLayout->addWidget(Label3, 0, 0);
	placeLayout->addWidget(Label4, 2, 0);
	placeLayout->addWidget(pb2,0, 1);
	placeLayout->addWidget(Label5,1, 1);
	placeLayout->addWidget(pb3,2, 1);
	Group2->setLayout(placeLayout);

	QVBoxLayout *mainLayout = new QVBoxLayout;
	mainLayout->addWidget(Label1);
	mainLayout->addWidget(Group1);
	mainLayout->addWidget(Label2);
	mainLayout->addWidget(Group2);
	mainLayout->addStretch(1);
	setLayout(mainLayout);
}

void PrivacyTab::setDefaultNek()
{
	nekLineEdit->setText(QString("HomePlug"));
}

void PrivacyTab::setLocalDevice()
{
	QString& my = Parent->getMyMac();

	if (my.isEmpty())
	{
		QMessageBox::critical(this, tr("plc"), tr("No Local Device found!\n"), QMessageBox::Ok);
	}
	else
	{
		Nek_state = NEK_SENT_LOCAL;
		plc_set_network_key(my, nekLineEdit->text());
	}
}

void PrivacyTab::setAllDevices()
{
	QString& my = Parent->getMyMac();

	if (my.isEmpty())
	{
		QMessageBox::critical(this, tr("plc"), tr("No Local Device found!\n"), QMessageBox::Ok);
	}
	else
	{
		Nek_state = NEK_SENT_FIRST;
		plc_set_network_key(my, nekLineEdit->text());
	}
}

void PrivacyTab::display_nek_confirm()
{
	switch(Nek_state) {
		case NEK_SENT_LOCAL:
			QMessageBox::information(this, tr("plc"), tr("Private Network name changed successfully!\n"), QMessageBox::Ok);
			Nek_state = NEK_IDLE;
			break;
		case NEK_SENT_FIRST:
			Nek_state = NEK_SENT_SECOND;
			QMessageBox::information(this, tr("plc"), tr("Private Network name changed successfully!\nChange for all devices now."), QMessageBox::Ok);
			plc_broadcast_key();
			break;
		case NEK_SENT_SECOND:
			QMessageBox::information(this, tr("plc"), tr("Private Network name changed successfully on all devices!\n"), QMessageBox::Ok);
			Nek_state = NEK_IDLE;
			break;
		default:
			QMessageBox::warning(this, tr("plc"), tr("Private Network name changed by somebody in the network!\n"), QMessageBox::Ok);
			break;
	}
}

AboutTab::AboutTab(plc *parent) : QTabWidget(parent)
{
	Parent = parent;
	QLabel *Label1 = new QLabel(tr("plc Version %1").arg(VERSION));
	QLabel *Label2 = new QLabel(tr("Released under GNU GPL v3.\n"));
	QLabel *Label3 = new QLabel(tr("Many thanks to:\n - Alexandre for the nice logo\n - Serge for his help on C++\n - Easyvisio for using GPL license.\n - Trolltech and Nokia for QT (QT Version : %1)").arg(QT_VERSION_STR));

	QLabel *Label4 = new QLabel(this);
	const QPixmap *logo = new QPixmap(":/plc_logo_small.png");
	Label4->setPixmap(*logo);

	QGroupBox *pref = new QGroupBox(tr("Preferences"));
	QCheckBox *AutoScan;
	AutoScan = new QCheckBox(tr("Auto Scan"));
	AutoScan->setChecked(true);
	connect(AutoScan, SIGNAL(toggled(bool)), this, SLOT(AutoScan_toggled(bool)));
	QVBoxLayout *prefLayout = new QVBoxLayout;
	prefLayout->addWidget(AutoScan);
	pref->setLayout(prefLayout);
	QVBoxLayout *mainLayout = new QVBoxLayout;
	mainLayout->addWidget(Label1);
	mainLayout->addWidget(Label2);
	mainLayout->addWidget(Label3);
	mainLayout->addWidget(Label4);;
	mainLayout->addWidget(pref);
	mainLayout->addStretch(1);
	setLayout(mainLayout);
}

void AboutTab::AutoScan_toggled(bool val)
{
	Parent->autoscan = val;
}

ScanThread::ScanThread(plc *par)
:QThread(par)
{
	p=par;
}

void ScanThread::run()
{
	while (! p->closing )
	{
		if (p->autoscan)
		{
			if ((p->getMyMac()).isEmpty())
			{
				plc_send_mm(); // find the adapter on witch we're attached
			}
			else
			{
				plc_get_channel_capacities(); // find connected (through Homeplug) devices and data rate
			}
		}
		else
		{
			qDebug("Autoscan off");
		}
		emit remove_old();
		sleep(AUTOSCAN_TEMPO);
	}
	qDebug("scanThread exited");
}

PcapThread::PcapThread(plc *parent)
:QThread(parent)
{
}

void PcapThread::run()
{
	plc *p = qobject_cast<plc *>(parent());
	qDebug() << "sniff thread start.";
	p->register_callback();
	qDebug() << "sniff thread end."; /* will probably never been reached since we're using terminate() to stop the thread */
}

void plc::change_if(const QString &argif)
{
	int res;
	qDebug() << "New device selected :" << argif;

	res = plcnet_select_if(QNetworkInterface::interfaceFromName (argif));
	if (res < 0)
	{
		qDebug() << "warning, device" << argif << "not supported";
	}
	else
	{
		qDebug() << "Stopping old sniff thread.";
		if (pcapThread != NULL)
			pcapThread->terminate();
		qDebug() << "start new sniff thread.";
		pcapThread = new PcapThread(this);
		pcapThread->start(QThread::NormalPriority);
	}
}

void plc::register_callback()
{
	qDebug() << "register_callback";
	plcnet_register_callback(StaticHandlePacket,reinterpret_cast<unsigned char*>(this));
	qDebug() << "register_callback ended";
}

//callback vue par le C
//static
void plc::StaticHandlePacket(unsigned char* instance, const struct pcap_pkthdr *header, const u_char *packet)
{
	plc* pThis = reinterpret_cast<plc*>(instance);
	return pThis->HandlePacket(header, packet);
}

//callback objet
void plc::HandlePacket(const struct pcap_pkthdr *header, const u_char *packet)
{
	const unsigned char *framebuf = packet;
	char macbuf[20];
	int i;
	unsigned int j;
	unsigned char mid;
	unsigned char melen, metype;
	const unsigned char *macmgmt;

	if (closing)
		return;

	if (header->len <= 0)
	{
		qDebug("packet len = %d", header->len);
		return;
	}
#if 0
	qDebug("Received packet !!!");
	for (i=0; i< header->len; i++)
		printf("%2.2X ", framebuf[i]);
	printf("\b \b\n");
#endif
	/* It's an Ethernet packet - read MAC source */
	QString NewMac(format_mac_addr(&framebuf[ETHER_ADDR_LEN], macbuf));
	emit Mac_detected(NewMac);

	melen = framebuf[16];
	metype = framebuf[15] & 0x1F; /* 5 bits */
	qDebug("From %s : 0x%X", format_mac_addr(&framebuf[6], macbuf), metype);
	switch (metype)
	{
		case 0x00:
			qDebug("Request Channel Estimation");
			break;
		case 0x01:
			qDebug("Channel Estimation response");
			break;
		case 0x02:
			qDebug("Vendor Specific Parameters");
			mid = framebuf[20] & 0x7F; // remove first bit (1 : response)
			if (!(framebuf[20] & 0x80))
			{
				qDebug() << "MID = " << mid << "(request)";
				return;
			}
			else
			{
				emit Date_detected(NewMac);
				qDebug() << "MID = " << mid << "(response)";
			}
			if ((framebuf[17] != 0x00) | (framebuf[18] != 0x04) | (framebuf[19] != 0x87))
			{
				qDebug("Warning, Vendor != Intellon");
			}

			switch (mid)
			{
				case 16:
					{
						QString macver = "MAC Version V";
						qDebug() << "Protocol Version Flag =" << (framebuf[21] & 0x02);
						if ((framebuf[22] == 0x55) && (framebuf[23] == 0))
						{
							macver += QString::number(framebuf[24],10) + "." + QString::number(framebuf[25],10);
						}
						else
						{
							macver += "1.2";
						}
						qDebug() << macver;
						if ((framebuf[21] & 0x01) == 0)
						{
							qDebug("Local Connection : %s", format_mac_addr(&framebuf[6], macbuf));
							emit My_Mac_detected(NewMac, macver);
						}
						else
						{
							qDebug("Remote Connection : %s", format_mac_addr(&framebuf[6], macbuf));
						}
						break;
					}
				case 24:
					if ( NewMac != getMyMac())
						break; // don't trace RX TX values of all devices

					switch (capabilities_flag) {
						unsigned char n;

						case CAPABILITIES_RX:
						qDebug("received channel capacities response mme for RX");

						if ((melen-4)%14 != 0)
						{
							qWarning("Error, ignore frame!");
							break;
						}
						n = (melen -4) / 14;
						qDebug("\t%d RX station(s)", n);
						for (i=0; i<n; i++)
						{
							qDebug("\tStation %d", i);
							qDebug("\t\tMAC %02x:%02x:%02x:%02x:%02x:%02x",
									framebuf[21+i*14],
									framebuf[22+i*14],
									framebuf[23+i*14],
									framebuf[24+i*14],
									framebuf[25+i*14],
									framebuf[26+i*14]);
							j = ex_word(&framebuf[27+i*14]);
							//qDebug("\t\tRxBytesPerBlock = %d", j);
							qDebug("\t\tRxBytesRate = %d MBytes/s", j/336);
							qDebug("\t\tRxBitRate   = %d MBits/s", 8*j/336);
							//qDebug("\t\tRxClockError = %d", *(unsigned int*)(&framebuf[29+i*14]));
							// ignore the 2 following bytes (reserved)

							//		if ( NewMac == p->getMyMac())
							{
								QString Mac(format_mac_addr(&framebuf[21+i*14], macbuf));
								emit Rate_detected(Mac, 8*j/336, 0);
							}
						}
						break;
						case CAPABILITIES_TX:
						qDebug("\treceived channel capacities response mme for TX");
						if ((melen-4)%8 != 0)
						{
							qWarning("Error, ignore frame !");
							break;
						}
						n = (melen -4) / 8;
						qDebug("\t%d TX station(s)", n);
						for (i=0; i<n; i++)
						{
							qDebug("\tStation %d", i);
							qDebug("\t\tMAC %02x:%02x:%02x:%02x:%02x:%02x",
									framebuf[21+i*8],
									framebuf[22+i*8],
									framebuf[23+i*8],
									framebuf[24+i*8],
									framebuf[25+i*8],
									framebuf[26+i*8]);
							j = ex_word(&framebuf[27+i*8]);
							qDebug("\t\tTxBytesPerBlock = %d", j);
							qDebug("\t\tTxBytesRate = %d MBytes/s", j/336);
							qDebug("\t\tTxBitRate   = %d MBits/s", 8*j/336);
							if ( NewMac == getMyMac())
							{
								QString Mac(format_mac_addr(&framebuf[21+i*8], macbuf));
								emit Rate_detected(Mac, 8*j/336, 1);
							}
						}
						break;
						default:
						qWarning("capabilities flag in bad mood");
						break;
					}
					break;
				default:
					qWarning("received unknow mid !");
					break;
			}
			break;
		case 0x03:
			qDebug("Replace Bridge Address");
			break;
		case 0x04:
			qDebug("Set Network Encryption Key");
			break;
		case 0x05:
			qDebug("Multicast with Response");
			break;
		case 0x06:
			qDebug("Confirm Network Encryption Key");
			emit Nek_changed_confirm();
			break;
		case 0x07:
			qDebug("Parameters and Statistics request");
			break;
		case 0x08:
			qDebug("Parameters and Statistics response");
			qDebug("\tlen = %d", framebuf[16]);
			macmgmt=&framebuf[15];
			qDebug()<< "\tTx ACK Counter:             " << ex_word(&macmgmt[2])
				<< "\tTx NACK Counter:            " << ex_word(&macmgmt[4])
				<< "\tTx FAIL Counter:            " << ex_word(&macmgmt[6])
				<< "\tTx Contention Loss Counter: " << ex_word(&macmgmt[8])
				<< "\tTx Collision Counter:       " << ex_word(&macmgmt[10])
				<< "\tTx CA3 Latency Counter:     " << ex_word(&macmgmt[12])
				<< "\tTx CA2 Latency Counter:     " << ex_word(&macmgmt[14])
				<< "\tTx CA1 Latency Counter:     " <<  ex_word(&macmgmt[16])
				<< "\tTx CA0 Latency Counter:     " << ex_word(&macmgmt[18])
				<< "\tRx Cumul. Bytes per 40-symbol Packet Counter:" << ex_long(&macmgmt[20]);
			break;
		case 0x12:
			qDebug("Get Device Version request, ignore");
			break;
		case 0x13:
			qDebug("Get Device Version response");
			if (framebuf[16] == 0x24)
			{
				qDebug("\tFirmware = %s", &framebuf[17]);
				QString firmware((char *)&framebuf[17]);
				emit My_Mac_detected(NewMac, firmware);
			}
			break;
		case 0x1a:		/* Intellon specific network statistics */
			if (!(framebuf[17] & 0x80)) 		/* Really a response? */
			{
				qDebug("YYYYYYYYYYYYYYYEEEEEEEEEEEEEESSSSSSSSSSSSSSS ! Intellon-specific network statistics");
			}
			else
			{
				qDebug("Intellon-specific network statistics request");
			}
			break;
		case 0x1c:
			qDebug("hip hip hip hourra !!!!");
			break;
		case 0x1f:
			qDebug("Set transmit characteristics");
			break;
		default:
			qDebug("Unknown response (METYPE = 0x%02x)", metype);
	}
}
